/*
 * $Id: ibargraph.c$
 *
 * LCDproc ISDN bargraph transfer rate client
 * based on code from imonlcd by Lars Wessels
 * which
 * includes code stolen from isdn4linux/imontty.c by Volker Gtz
 *
 * Copyright (c) 2003 Hendrik Lipka <ibargraph@hendriklipka.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 *
 * $Log: ibargraph.c $
 * Revision 0.1
 * Initial revision
 * - base on imonlcd.c
 * - bugfix for buffer overflow
 * - bugfixes for memory leak
 * - handles single ISDN channel
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <linux/isdn.h>

#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>

#include "sockets.h"
#include "ibargraph.h"
#include "debug.h"

FILE *isdninfo;
int sock=0;
int verbose=0;
channel_info info[ISDN_MAX_CHANNELS];
int transfer_old[ISDN_MAX_CHANNELS];
int transfer_current[ISDN_MAX_CHANNELS];

static void get_if_state();
static char *home;

static struct timeval timeout, tv_last, tv_up;
static int usageflags[ISDN_MAX_CHANNELS];
static Siobytes iobytes[ISDN_MAX_CHANNELS];
static unsigned long in_bytes_last, out_bytes_last;
static unsigned long in_bytes_per_sec, out_bytes_per_sec;
static unsigned long channels_last;

static int filter=4;

void usage() {
	printf("%s v%s - (c) 2003 Hendrik Lipka <ibargraph@hendriklipka.de>\n",
		CLIENTNAME, VERSION);
	fprintf(stderr,"Usage: ibargraph [-s server] [-p port] [-v]\n");
	exit(0);
}

void scan_str(char *buffer, char *field[], int max) {

	char *buf = buffer;
	int i;
	char *token = strtok(buf, "\t");

	for(i=0; i < max; i++) {
		token = strtok(NULL, " ");
		if (token == NULL) continue;
		if ((field[i] = malloc(strlen(token) * sizeof(char))) != NULL)
			strcpy(field[i], token);
	}
}

/*
    frees all memory allocted by scan_str
*/
int free_str(char *field[], int max)
{
    int i;
    for(i=0; i < max; i++)
    {
        if (NULL!=field[i])
        {
            free(field[i]);
            field[i]=NULL;
        }
    }
    return 0;
}

void scan_int(char *buffer, int (*field)[], int max) {

	char *buf = buffer;
	int i;
	char *token = strtok(buf, "\t");

	for(i=0; i < max; i++) {
		token = strtok(NULL, " ");
		if (token == NULL) continue;
		(*field)[i] = atoi(token);
	}
}

void clean_exit (int val) {
    sock_close(sock);
    fclose(isdninfo);
    exit(0);
}

void alarm_exit (int val) {
	fprintf(stderr,"No answer from LCDproc server. Exiting.\n");
   	clean_exit(0);
}

/* get channel info from /dev/isdninfo, inspired by imontty.c */
int read_isdn_info (FILE *isdninfo, channel_info *info) {

    char buf[BUFSIZE];
    char *idmap[ISDN_MAX_CHANNELS];
    int chmap[ISDN_MAX_CHANNELS];
    int drmap[ISDN_MAX_CHANNELS];
    int usage[ISDN_MAX_CHANNELS];
    int flags[ISDN_MAX_DRIVERS];
    char *phone[ISDN_MAX_CHANNELS];
    int i, active=0;


    fgets(buf, BUFSIZE, isdninfo);
    scan_str(buf, idmap, ISDN_MAX_CHANNELS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_int(buf, &chmap, ISDN_MAX_CHANNELS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_int(buf, &drmap, ISDN_MAX_CHANNELS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_int(buf, &usage, ISDN_MAX_CHANNELS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_int(buf, &flags, ISDN_MAX_DRIVERS);

    fgets(buf, BUFSIZE, isdninfo);
    scan_str(buf, phone, ISDN_MAX_CHANNELS);


    for(i=0; i < ISDN_MAX_CHANNELS; i++, info++) {
		if (chmap[i] >= 0 && drmap[i] >= 0) {
			active++;

			info->chmap = chmap[i];
			info->idmap = idmap[i];

			/* values for usage inspired by display's character set */
			if (usage[i] & ISDN_USAGE_OUTGOING)
				info->usage = 126;  /* char value for "->" */
			else if (usage[i] == ISDN_USAGE_EXCLUSIVE)
				info->usage = 69;   /* char value for "E"  */
			else if (usage[i])
				info->usage = 127;  /* char value for "<-" */
			else
				info->usage = 79;   /* char value for "O"  */

			switch(usage[i] & ISDN_USAGE_MASK) {
				case ISDN_USAGE_RAW:
					info->type = "Raw"; break;
				case ISDN_USAGE_MODEM:
					info->type = "Modem"; break;
				case ISDN_USAGE_NET:
					info->type = "Net"; break;
				case ISDN_USAGE_VOICE:
					info->type = "Voice"; break;
				case ISDN_USAGE_FAX:
					info->type = "Fax"; break;
				default:
					info->type = "?"; break;
			}

			if(usage[i] & ISDN_USAGE_MASK)
				info->phone = phone[i];
		}
	}
	/* now free all memory allocated by scan_str */
	free_str(idmap, ISDN_MAX_CHANNELS);
    free_str(phone, ISDN_MAX_CHANNELS);
	return active;
}


/**/
int check_isdn_status (FILE *isdninfo, channel_info *info) {

	struct timeval timeout;
	fd_set fdset;

	FD_ZERO(&fdset);
	FD_SET(fileno(isdninfo),&fdset);
	timeout.tv_sec =  0;
	timeout.tv_usec = 250;

    get_if_state();

	switch (select(32,&fdset,(fd_set *)0,(fd_set *)0,&timeout)) {
		case 0: /* nothing changed */
			return 0;
		case 1:
			read_isdn_info(isdninfo, info); /* read new ISDN info */
			return 1;
		default:
			fprintf(stderr,"ERROR: select(%s): %s\n",
					ISDN_INFO_DEV, strerror(errno));
			clean_exit(1);
	  }
}

/* setup initial ISDN line monitor screen
   returns number of active ISDN channels  */
int init_screen (void) {

	int i, active;
	char tmp[MAXMSG];
	channel_info *ch;

	sprintf(tmp, "client_set name {%s}\n",CLIENTNAME);
	sock_send_string(sock, tmp);
	sprintf(tmp, "output 0\n");
	sock_send_string(sock, tmp);

	/* get initial info about available ISDN channels */
	active = read_isdn_info(isdninfo, info);

	return active;
}

int parse_server_reply(char *buffer, char *field[]) {

	char *reply_token, *cp, *xx;
	int i,tokens,len,mlen;
//    printf("%s\n",buffer);
	for(i=0,tokens=0; i < MAXTOKENS; i++)
	{
		if (i==0)
		    reply_token = strtok(buffer, " ");
		else
		    reply_token = strtok(NULL, " ");
		if (reply_token == NULL)
		    continue;
		len=strlen(reply_token);

        mlen=len*sizeof(char);
		//printf("len=%i\n",len);
		//printf("mem len=%i\n",mlen);
		//printf("i=%i\n",i);

		xx=malloc(mlen+1);
//		printf("malloc %i =%i\n",i,xx);
//		printf("state=%i\n",mprobe(xx));
		field[i]=xx;
		if (field[i] != NULL)
		{
			if ((cp = strchr(reply_token, '\n')) != NULL)
			    *cp = '\0';
			strncpy(field[i], reply_token,mlen+1);
			//printf("%s\n",field[i]);
			tokens++;
		}
//		printf("state=%i\n",mprobe(xx));
	}
	return tokens;
}

/*
    frees all memory allocted by parse_server_reply
*/
int free_reply(char *field[])
{
    int i;
    for(i=0; i < MAXTOKENS; i++)
    {
        if (NULL!=field[i])
        {
//		    printf("free %i=%i\n",i,field[i]);
            free(field[i]);
            field[i]=NULL;
        }
    }
    return 0;
}

int main(int argc, char **argv) {

	char server[BUFSIZE] = "localhost";
	int port = LCDPORT;
	int i, j, p, x, tokens, ch_status, ch_status_prev=0,
		len=0, ch_active=0, sendinfo=0;
	channel_info *ch;
	char tmp[MAXMSG], buf[MAXMSG], *reply[MAXTOKENS];
	int transfer_rate;

	/* install signal handlers */
	signal(SIGINT, clean_exit);
	signal(SIGHUP, clean_exit);
	signal(SIGTERM, clean_exit);
	signal(SIGALRM, alarm_exit);

    /* parse command line */
	for(i=1, j=0; i<argc; i++) {
		if(argv[i][0] == '-') {
			switch(argv[i][1]) {
				/* server name */
				case 'S':
				case 's': if(argc < i+1) usage();
					strcpy(server, argv[++i]);
					 break;
				/* port */
				case 'P':
				case 'p': if(argc < i+1) usage();
					port = atoi(argv[++i]);
					break;
				/* verbose outpu */
				case 'V':
				case 'v': verbose=1;
					break;
				default: usage();
					break;
	    	}
		}
	}

	if (!(isdninfo = fopen(ISDN_INFO_DEV, "r"))) {
		fprintf(stderr, "ERROR: can't open '%s'\n",
				ISDN_INFO_DEV, strerror(errno));
		exit(1);
	}

	/* connect to the LCDd on server */
	sock = sock_connect(server, port);
	if (sock <= 0) {
		fprintf(stderr,"ERROR: can't connect to server %s on port %i.\n",
			 server, port);
		clean_exit(1);
	}

	/* initial client->server handshake with timeout */
	sock_send_string(sock, "hello\n");
	alarm(15);
	while (len <= 0) {
		len = sock_recv(sock, buf, MAXMSG);
	}
	alarm(0);
	debug("socket in: %s\n", buf);

	/* is there an LCDproc server on the other end? */
	tokens = parse_server_reply(buf, reply);
	if (tokens > 0 && !strcmp(reply[0], "connect")) {
		if (verbose)
			printf("connected to LCDproc server %s\n", reply[2]);
	} else {
		fprintf(stderr,"ERROR: handshaking with server failed:\n");
		fprintf(stderr,"expecting \"connect\" ");
		fprintf(stderr,"but server says \"%s\"\n", buf);
		clean_exit(1);
	}
	memset(&buf,0,sizeof(buf));

	/* init screen widgets and get number of active ISDN channels */
	ch_active = init_screen();

	/* main update loop */
	while(1) {

		/* pay attention to server commands */
		len = sock_recv(sock, buf, MAXMSG);
		if (len > 0) {
			debug("socket in: %s\n", buf);
			tokens = parse_server_reply(buf, reply);
			memset(&buf,0,sizeof(buf));

			/* is he talking to us ? */
			if (tokens > 1 && !strcmp(reply[1], CLIENTNAME))
				sendinfo = strcmp(reply[0], "listen") ? 0 : 1;

			/* server terminates */
			if (!strcmp(reply[0], "bye")) {
				fprintf(stderr,"LCDproc server said 'bye'. Exiting.\n");
				clean_exit(0);
			}
		} else if (!len) {
			fprintf(stderr,"LCDproc probably died. Exiting.\n");
			clean_exit(0);
		}
	    /* now free all memory allocated by parse_server_reply */
        free_reply(reply);

        /* get ISDN status information */
		ch_status = 0;
		check_isdn_status(isdninfo, info);

        /* get transfer rate maximum */
		if (in_bytes_per_sec>out_bytes_per_sec)
		    transfer_rate=in_bytes_per_sec;
		else
		    transfer_rate=out_bytes_per_sec;
        /* set bar graph */
        setBargraphState(transfer_rate, channels_last);


		usleep(300000);
	}
}

static void get_if_state()
{
    static char buf[4096];
    struct timeval tv_now;
    int    ms_delta;
    int    in_bytes_now, out_bytes_now;
    int  fd;
    int  i;

    /* Open the info device */
    if ((fd = open(ISDN_INFO_DEV, O_RDONLY | O_NDELAY)) < 0)
    {
        syslog(LOG_ERR, "%s: %s\n", ISDN_INFO_DEV, sys_errlist[errno]);
	    closelog();
	    exit(1);
    }

    /* Whats the time now */
    gettimeofday(&tv_now, NULL);
    ms_delta = (tv_now.tv_sec * 1000 + tv_now.tv_usec / 1000) -
        (tv_last.tv_sec * 1000 + tv_last.tv_usec / 1000);
    tv_last = tv_now;

    /* Get info from interface */
    if (read(fd, buf, sizeof(buf))> 0)
    {
    	sscanf(strstr(buf, "usage:"),
    	"usage: %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
    	&usageflags[0], &usageflags[1], &usageflags[2], &usageflags[3],
    	&usageflags[4], &usageflags[5], &usageflags[6], &usageflags[7],
    	&usageflags[8], &usageflags[9], &usageflags[10], &usageflags[11],
    	&usageflags[12], &usageflags[13], &usageflags[14], &usageflags[15]);
    }
	in_bytes_now = 0;
	out_bytes_now = 0;
	/* Get byte in/out for all channels */
	if (ioctl(fd, IIOCGETCPS, &iobytes))
	{
    	syslog(LOG_ERR, "%s: %s\n", IIOCGETCPS, sys_errlist[errno]);
    	closelog();
    	exit(1);
	}
	close(fd);

    channels_last=0;
	/* Count number of open channes and total in/out bytes */
	for (i = 0; i < ISDN_MAX_CHANNELS; i++)
	{
    	if (usageflags[i])
    	{
        	in_bytes_now += iobytes[i].ibytes;
        	out_bytes_now += iobytes[i].obytes;
        	channels_last++;
    	}
	}

	/* Calculate the total through put in bytes/sec */
	if (filter < 1)
	{
    	in_bytes_per_sec = (in_bytes_now - in_bytes_last) * 1000 / ms_delta;
    	out_bytes_per_sec = (out_bytes_now - out_bytes_last) * 1000 / ms_delta;
	}
	else
	{
    	in_bytes_per_sec = (in_bytes_per_sec * (filter - 1) +
    	    (in_bytes_now - in_bytes_last) * 1000 / ms_delta) / filter;
    	out_bytes_per_sec = (out_bytes_per_sec * (filter - 1) +
    	    (out_bytes_now - out_bytes_last) * 1000 / ms_delta) / filter;
	}
	in_bytes_last = in_bytes_now;
	out_bytes_last = out_bytes_now;
}

/*
    sets the bar graph for the transfer rate
*/
int setBargraphState(int bytes_per_second, int channels)
{
    char tmp[MAXMSG];

    /***********************************************************************/
    /**** CHANGE THIS TO YOUR DISPLAY !! ***********************************/
    /***********************************************************************/
    /*  normal bar */
    /* int bar[]={0,1,3,7,15,31,63,217,255}; */

    /*  normal bar, but inverted port */
    /* int bar[]={255,254,252,255-7,255-15,255-31,255-63,255-217,0}; */

    /* bar upside down */
    /* int bar[]={0,128,64+128,32+64+128,16+32+64+128,8+16+32+64+128,4+8+16+32+64+128,2+4+8+16+32+64+128,255}; */

    /* bar upside down, inverted port */
    /* int bar[]={255,127,63,31,15,7,3,1,0}; */

    /* the bar itself is upside down, and the bits 0 and 2 are swapped :( */
    /* I really should learn soldering :-| */
    int bar[]={0,128,64+128,32+64+128,16+32+64+128,8+16+32+64+128,1+8+16+32+64+128,2+1+8+16+32+64+128,255};

    int graph;
    int transfer;

    if (0==channels)
    {
        transfer=0;
        channels=1;
    }
    transfer=bytes_per_second/(1000*channels);

    if (transfer>8)
        transfer=8;
    graph=bar[transfer];
//    printf("rate / bar / graph: %i/%i/%i\n",bytes_per_second,transfer,graph);

    sprintf(tmp, "output %i\n",graph);

	sock_send_string(sock, tmp);

    return 0;
}
    